void pruneTreeDensityLevel2DTI(double **y, int m, int n, double gamma, 
		   double delta, double penWeight, double **S, char *noiseType)
{
  int L,N, in,im,iL,j,kn,km,iL2, dyadLen;
  double ***SScaling;
  double **Offset, **Offset_prev, **cost, **cost_prev;
  double ***waveletScale, ***splitDecision;
  double xD1, xD2, xD3, xD4;
  double Cu, Cl, qA, pA, volA, mergeCost, splitCost;
	
  L = (int) floor(log(min(m,n))/log((double)2));
	
  Offset = (double **)mxCalloc(m,sizeof(double*));
  Offset_prev = (double **)mxCalloc(m,sizeof(double*));
  cost = (double **)mxCalloc(m,sizeof(double*));
  cost_prev = (double **)mxCalloc(m,sizeof(double*));
	
  SScaling = (double ***)mxCalloc(m,sizeof(double**));
  waveletScale = (double ***)mxCalloc(m,sizeof(double**));
  splitDecision = (double ***)mxCalloc(m,sizeof(double**));

  N = 0;	
  for (im = 0; im < m; im++) {
    Offset[im] = (double *)mxCalloc(n,sizeof(double));
    Offset_prev[im] = (double *)mxCalloc(n,sizeof(double));
    cost[im] = (double *)mxCalloc(n,sizeof(double));
    cost_prev[im] = (double *)mxCalloc(n,sizeof(double));
		
    SScaling[im] = (double **)mxCalloc(n,sizeof(double*));
    waveletScale[im] = (double **)mxCalloc(n,sizeof(double*));
    splitDecision[im] = (double **)mxCalloc(n,sizeof(double*));
		
    for (in = 0; in <n; in++) {
      SScaling[im][in] = (double *)mxCalloc(L+1,sizeof(double));
      waveletScale[im][in] = (double *)mxCalloc(L,sizeof(double));
      splitDecision[im][in] = (double *)mxCalloc(L,sizeof(double));
      N += y[im][in];
    }
  }
  /* initialize variables */
  qA = 3*L+2;
  volA = 1.0/m/n;
  pA = 4*max(1, (log(1/delta)+qA*log(2)))/N;
  /*printf("init Tree = \n");*/
  for (im = 0; im < m; im++) {
    for (in = 0; in < n; in++){
      Offset_prev[im][in] = (gamma*volA-y[im][in]/N)/2;
      S[im][in] = (double) (Offset_prev[im][in]<0);
      cost_prev[im][in] = Offset_prev[im][in]*(2*S[im][in]-1)+
		penWeight*pow(2*(log(2/delta)+qA*log(2))/N*
		      (pA*(1-S[im][in])+gamma*volA*S[im][in]),0.5);
      SScaling[im][in][0] = S[im][in];
    }
  }
	
  /*printf("All set: %f\t%f\t%f\n",gamma,delta,penWeight);*/
  for (iL=0;iL<L;iL++) {
    dyadLen = (int)pow(2.0,iL+1);
    /*printf("\ndyadLen = %d\n",dyadLen);*/
    qA = 3*(L-iL)+2;
    volA = dyadLen*dyadLen/m/n;
    /*pA = 4*max(pow(dyadLen,2.0), (log(1/delta)+qA*log(2)))/m/n;*/
    pA = pow(dyadLen,2.0)/N;
    for (im = 0; im < m; im++) {
      for (in = 0; in < n; in++){
        km = (int)mod(im+dyadLen/2,m);
        kn = (int)mod(in+dyadLen/2,n);
				
	splitCost = cost_prev[im][in]+cost_prev[km][in]+
	  cost_prev[im][kn]+cost_prev[km][kn];
				
	Offset[im][in] = Offset_prev[im][in]+Offset_prev[km][in]+
	  Offset_prev[im][kn]+Offset_prev[km][kn];
				
        /* calculate current scaling coefficient */
        /*SScaling[im][in][iL+1] = (SScaling[im][in][iL]+SScaling[im][kn][iL]+
	  SScaling[km][in][iL]+SScaling[km][kn][iL])/4.0;*/
	SScaling[im][in][iL+1] = (Offset[im][in]<0); 

	mergeCost = Offset[im][in]*(2*(Offset[im][in]<0)-1)+
	  penWeight*pow(2*(log(2/delta)+qA*log(2))/N*
			(pA*(Offset_prev[im][in]>=0) + 
			 gamma*volA*(Offset_prev[im][in]<0))
			,0.5);
				
	/*printf("SScaling[%d][%d][%d]=%f\t",im,in,iL+1,SScaling[im][in][iL+1]*15);*/
				
        /* decide whether to split or merge, save decision and associated
	   log probability */
        if (splitCost > mergeCost ) {
	  cost[im][in] = mergeCost;
          splitDecision[im][in][iL] = 0.0;
        }
        else {
	  cost[im][in] = splitCost;
          splitDecision[im][in][iL] = 1.0;
        }
        waveletScale[im][in][iL] = 1.0;
      }
    }
		
		
    /* set up scaling coefficients for next scale */
    for (im = 0; im < m; im++) {
      for (in = 0; in < n; in++) {
	cost_prev[im][in] = cost[im][in];
	Offset_prev[im][in] = Offset[im][in];
      }
    }
  }
	
  for (iL=L-1; iL>=0; iL--) {
    dyadLen = (int)pow(2.0,iL+1);
    /*printf("\ndyadLen = %d, L = %d\n",dyadLen,L);*/
		
    for (in = 0; in < n; in++) {
      for (im = 0; im < m; im++) {
        km = (int)mod(im+dyadLen/2,m);
        kn = (int)mod(in+dyadLen/2,n);
        /* if the split decision is zero, then the associated wavelet should
	   be set to zero. */
        if (splitDecision[im][in][iL]==0) {
          waveletScale[im][in][iL] = 0.0;
        }
	/*printf("waveletScale[%d][%d][%d]=%f\n",im,in,iL,waveletScale[im][in][iL]);*/
				
        if (iL > 0) {
          /* threshold wavelet coefficient based on ancestors' split/merge
	     decisions */
          waveletScale[im][in][iL-1] = waveletScale[im][in][iL-1]
	    - (0.25*(1.0-waveletScale[im][in][iL]));
          waveletScale[im][kn][iL-1] = waveletScale[im][kn][iL-1]
            - (0.25*(1.0-waveletScale[im][in][iL]));
          waveletScale[km][in][iL-1] = waveletScale[km][in][iL-1]
            - (0.25*(1.0-waveletScale[im][in][iL]));
          waveletScale[km][kn][iL-1] = waveletScale[km][kn][iL-1]
            - (0.25*(1.0-waveletScale[im][in][iL]));
        }
      }
    }
		
    /* construct estimate based on wavelet coefficients and thresholds */
    for (im = 0; im < m; im++) {
      for (in = 0; in < n; in++) {
	kn = (int)mod(in-dyadLen/2,n);
        km = (int)mod(im-dyadLen/2,m);
				
	/*printf("iL=%d\tdyadLen=%d\t[im,in]=[%d,%d]\t[km,kn]=[%d,%d]\n",iL,dyadLen,im,in,km,kn);*/
	/*printf("waveletScale[%d][%d][%d]=%f\n",iL,im,in,waveletScale[im][in][iL]);*/
        xD1 = SScaling[im][in][iL]*waveletScale[im][in][iL]+S[im][in]*(1.0-waveletScale[im][in][iL]);
        xD2 = SScaling[im][in][iL]*waveletScale[im][kn][iL]+S[im][kn]*(1.0-waveletScale[im][kn][iL]);
        xD3 = SScaling[im][in][iL]*waveletScale[km][in][iL]+S[km][in]*(1.0-waveletScale[km][in][iL]);
        xD4 = SScaling[im][in][iL]*waveletScale[km][kn][iL]+S[km][kn]*(1.0-waveletScale[km][kn][iL]);

	Offset[im][in] = (xD1 + xD2 + xD3 + xD4)/4.0;
      }
    }
    /*printf("iL=%d, Scaling Coeffs = \n",iL);*/
    for (im = 0; im < m; im++) {
      for (in = 0; in < n; in++) {
        S[im][in] = Offset[im][in];
	/*printf("%f\t",S[im][in]);*/
      }
      /*printf("\n");*/
    }
    
		
  }
  for (im = 0; im < m; im++) {
    for (in = 0; in < n; in++) {
      S[im][in] = (S[im][in]>0.5);
    }
  }
}
